/* Copyright (C) Eliot Stock - All Rights Reserved
 * Unauthorized copying of this file, via any medium is strictly prohibited.
 * Proprietary and confidential.
 * Written by Eliot Stock, 1@eliotstock.com
 */
#include <stdint.h>
#include <stdio.h>
#include <stdbool.h>

#include "app_error.h"
#include "nrf_delay.h"
#include "nrf_nvic.h"
#include "nrf_soc.h"

#include "config.h"
#include "board.h"
#include "debug_leds.h"
#include "error.h"

// From main.c
extern bool m_soft_device_enabled;

// We have three types of errors we need to handle in our application:
//   1. Asserts. These are when a call to an SDK function or one of our own returns something other
//      than success (0).
//   2. Soft device faults. These are probably caused by us misusing the soft device and indicate a
//      bug in our application code.
//   3. Hardfaults. If these are caused by our code, eg. dividing by zero, we should be able to fix
//      them. But they may also be caused by electromagnetic interference near the device or
//      glitches on power lines. We must handle them in production. We use the SDK's implementation
//      of the hardfault handler.

// Reset, but set the general purpose retention register first.
void app_reset(void)
{
	uint32_t err_code = 0;

	debug_led_off_all();

    if (m_soft_device_enabled)
    {
    	err_code = sd_power_gpregret_set(START_FROM_ASSERT);
    	APP_ERROR_CHECK(err_code);

    	err_code = sd_nvic_SystemReset();
    	APP_ERROR_CHECK(err_code);
    }
    else
    {
    	NRF_POWER->GPREGRET = START_FROM_ASSERT;
    	NVIC_SystemReset();
    }
}

// To pass in a file and line number, use:
//     app_error_handler(err_code, (uint16_t)__LINE__, (uint8_t *)__FILE__);
void app_error_handler(uint32_t error_code, uint32_t line_num, const uint8_t * p_file_name)
{
	debug_led_on(PIN_LED_ASSERT);

	printf("Error: %d %s:%d\n", (int)error_code, (char *)p_file_name, (int)line_num);

	// Copied in from <SDK>/softdevice/s130/headers/nrf_error.h. Update this list on each SDK release.
	switch (error_code)
	{
		case NRF_ERROR_SVC_HANDLER_MISSING:
			printf("NRF_ERROR_SVC_HANDLER_MISSING\n");
			break;
		case NRF_ERROR_SOFTDEVICE_NOT_ENABLED:
			printf("NRF_ERROR_SOFTDEVICE_NOT_ENABLED\n");
			break;
		case NRF_ERROR_INTERNAL:
			printf("NRF_ERROR_INTERNAL\n");
			break;
		case NRF_ERROR_NO_MEM:
			printf("NRF_ERROR_NO_MEM\n");
			break;
		case NRF_ERROR_NOT_FOUND:
			printf("NRF_ERROR_NOT_FOUND\n");
			break;
		case NRF_ERROR_NOT_SUPPORTED:
			printf("NRF_ERROR_NOT_SUPPORTED\n");
			break;
		case NRF_ERROR_INVALID_PARAM:
			printf("NRF_ERROR_INVALID_PARAM\n");
			break;
		case NRF_ERROR_INVALID_STATE:
			printf("NRF_ERROR_INVALID_STATE\n");
			break;
		case NRF_ERROR_INVALID_LENGTH:
			printf("NRF_ERROR_INVALID_LENGTH\n");
			break;
		case NRF_ERROR_INVALID_FLAGS:
			printf("NRF_ERROR_INVALID_FLAGS\n");
			break;
		case NRF_ERROR_INVALID_DATA:
			printf("NRF_ERROR_INVALID_DATA\n");
			break;
		case NRF_ERROR_DATA_SIZE:
			printf("NRF_ERROR_DATA_SIZE\n");
			break;
		case NRF_ERROR_TIMEOUT:
			printf("NRF_ERROR_TIMEOUT\n");
			break;
		case NRF_ERROR_NULL:
			printf("NRF_ERROR_NULL\n");
			break;
		case NRF_ERROR_FORBIDDEN:
			printf("NRF_ERROR_FORBIDDEN\n");
			break;
		case NRF_ERROR_INVALID_ADDR:
			printf("NRF_ERROR_INVALID_ADDR\n");
			break;
		case NRF_ERROR_BUSY:
			printf("NRF_ERROR_BUSY\n");
			break;
		case NRF_ERROR_CONN_COUNT:
			printf("NRF_ERROR_CONN_COUNT\n");
			break;
		case NRF_ERROR_RESOURCES:
			printf("NRF_ERROR_RESOURCES\n");
			break;
	}

	// If we end up here often enough in field testing, consider using ble_debug_assert_handler() from the SDK to write the stack to flash. We can then read it back on the desk. The alternative is to get gdb connected without resetting when we get back to the desk.

#if defined(DEBUG_NRF)
    // Generate breakpoint if debugger is connected. gdb will wait here for us.
	// TODO: This define isn't right. This line is being reached when the debugger's not connected and causing a reset right here.
    // __BKPT(0);
#endif

	// Wait a bit before we reset. In field testing, we need some time to see the assert LED go on.
    // TODO: Once we have preprocessor macros for a production build, leave this out of it.
	nrf_delay_ms(10 * 1000);

	// For a demo, just half a second and we might get away with an assert mid-demo:
	// nrf_delay_ms(500);

	app_reset();
}

void app_error_fault_handler(uint32_t id, uint32_t pc, uint32_t info)
{
	// TODO: When we get more debug LEDs, consider using one of them especially for faults.
	debug_led_on(PIN_LED_ASSERT);

	printf("Fault: id: %d pc: %d info: %d\n", (int)id, (int)pc, (int)info);

	// As of s130 v 2.0.0, the only fault ID we ever expect to get here is:
	// NRF_FAULT_ID_SD_ASSERT: The SoftDevice has triggered an assertion. Record the value of the
	// pc parameter and make it available to the Nordic support team to start an internal
	// investigation.

#if defined(DEBUG_NRF)
    // Generate breakpoint if debugger is connected. gdb will wait here for us.
    __BKPT(0);
#endif

	// Wait a bit before we reset. In field testing, we need some time to see the assert LED go on.
    // TODO: Once we have preprocessor macros for a production build, leave this out of it.
	nrf_delay_ms(10 * 1000);

	// For a demo, just half a second and we might get away with an assert mid-demo:
	// nrf_delay_ms(500);

	app_reset();
}

// Soft Device code calls this on errors (not faults).
void assert_nrf_callback(uint16_t line_num, const uint8_t * p_file_name)
{
    app_error_handler(DEAD_BEEF, line_num, p_file_name);
}
